/*
 * Copyright (c) 1996-1998 Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the SPAND Research
 *      Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/* $Header: /home/stemm/cvs/JavaSpand/spand/GetPerf.java,v 1.9 1998/06/27 03:23:51 stemm Exp $ */
package spand;
import java.net.*;
import java.util.*;
import java.lang.*;
import util.*;

/**
 * This is a SpandHandler object that calculates the typical
 * performance for a given address. The handler gets all reports for
 * an address and calculates the median bandwidth and round trip time
 * measurement. Clients query the performance for an address by
 * creating a message with "spand.GetPerf" as the handler and with a
 * PerfRequestPayload object as the payload. A PerfResponsePayload
 * object is returned.
 * @version $Revision: 1.9 $
 * @author SPAND Group, UC Berkeley.
 */

public class GetPerf implements SpandHandler {

  public Msg execute (Msg m, Hashtable globalHash){
    Msg reply = null;
    Hashtable reportHash=(Hashtable) globalHash.get("spand.ReportPayload");

    /* Make the reportHash if it doesn't yet exist. */
    if (reportHash==null){
      reportHash = new Hashtable();
      globalHash.put(new String("spand.ReportPayload"), reportHash);
    }

    /* Unmarshal the data portion of the message into a PerfRequestPayload object */
    PerfRequestPayload req = new PerfRequestPayload();

    req.Unmarshal(m.data, 0);

    /* Get the vector of Performance Reports for the specified address */
    Vector reports = (Vector) reportHash.get(req.addr);

    if (reports==null){
      reply = new Msg("spand.ErrorPayload",
			  new ErrorPayload(SpandError.NoReportsForAddr), false);
    } else {
      /* Make a new PerfResponsePayload */
      PerfResponsePayload p = CalcPerf(req.addr, reports);
      reply = new Msg("spand.PerfResponsePayload", p, false);
    }
    return reply;
  }

  private PerfResponsePayload CalcPerf(InetAddress addr, Vector reports){
    PerfResponsePayload perf = new PerfResponsePayload(addr);

    double sx, sy, ex, ey, sdx, sdy;
    ReportPayload r;
    double []bws;
    double []srtts;
    int i;
	
    sx=sy=0;

    bws= new double[reports.size()];
    srtts=new double[reports.size()];

    perf.numBytes=0;
    
    /* Loop over Report Payloads */
    Enumeration coll = reports.elements();
    r=(ReportPayload) coll.nextElement();

    for (i=0; coll.hasMoreElements(); r=(ReportPayload) coll.nextElement(), i++){

      perf.numBytes+=r.sampLen;
      double bits = r.sampLen*8;
      double sec = (double) r.sampDur/(double) 1000;
      double bw = (double) ((double) bits)/((double) sec);
      bws[i]=bw;
      srtts[i]=r.srtt;
      perf.numReports++;
      sx+=bw;
      sy+=r.srtt;
    }

    /* Loop again to calculate the standard deviation. We do this to
       avoid overflows while calculating the sum of squares */
    ex=sx/perf.numReports;
    ey=sy/perf.numReports;

    coll = reports.elements();
    r=(ReportPayload) coll.nextElement();

    sdx=sdy=0;
    for (i=0; coll.hasMoreElements(); r=(ReportPayload) coll.nextElement(), i++){
      double bits = r.sampLen*8;
      double sec = (double) r.sampDur/(double) 1000;
      double bw = (double) ((double) bits)/((double) sec);
      sdx+=(ex-bw)*(ex-bw);
      sdy+=(ey-r.srtt)*(ey-r.srtt);
    }

    /* Sort the array of bws and return the median as the answer */
    QSortAlgorithm q = new QSortAlgorithm();
    
    q.sort(bws);
    perf.availBandwidth=(int) bws[perf.numReports/2];
    
    /*    perf.stdDevAvailBandwidth=(int) Math.sqrt((double) (sx2/(double) perf.numReports)-((sx/(double) perf.numReports)*(sx/(double) perf.numReports)));*/
    perf.stdDevAvailBandwidth=(int) Math.sqrt(sdx/(perf.numReports));
    
    /* Sort the array of srtts and return the median as the answer */
    q.sort(srtts);
    perf.avgSrtt=(int) srtts[perf.numReports/2];
    
    /*    perf.stdDevSrtt=(int) Math.sqrt((double) (sy2/(double) perf.numReports)-((sy/(double) perf.numReports)*(sy/(double) perf.numReports)));*/

    perf.stdDevSrtt=(int) Math.sqrt(sdy/(perf.numReports));
    
    
    return perf;
  }

  /** Print out the data portion of the specified message */
  public String toString(Msg m){
    PerfRequestPayload s = new PerfRequestPayload();
    s.Unmarshal(m.data,0);

    return (s.toString());
  }
    
}
