/*
 * Copyright (c) 1996-1998 Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the SPAND Research
 *      Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/* $Header: /home/stemm/cvs/JavaSpand/spand/Listener.java,v 1.8 1998/06/26 16:29:16 stemm Exp $ */
package spand;
import java.net.*;
import java.io.*;
import java.util.*;

/**
 * A Listener is a thread that listens for SPAND Messages from
 * Clients. Every time a Message is received, the Handler specified in
 * the message is invoked, and if the Handler generates a reply
 * message, the reply is returned to the sender. A Listener can be
 * created that listens either on TCP or UDP ports. Listeners are
 * multi-threaded; they create a new thread for each incoming message.
 * @version $Revision: 1.8 $
 * @author SPAND Group, UC Berkeley.  */

public class Listener extends Thread{

    /** Used to listen for Datagram Requests */
    DatagramSocket requestDgramSock;
  
    /** Used to listen for Stream Requests */
    ServerSocket requestSock;

    /** Holds the active Stream Connection */
    Socket activeConn;

    /** The database of Performance Reports. The hash maps from String
      objects (which indicate the type of Performance Report) to
      arbitrary Java Objects. The SpandHandlers that use a particular
      Performance Report will know what the type of the value in the
      hash is. */
    Hashtable globalHash;

    /** If true, then use a stream socket to listen for SPAND Messages. */
    final boolean stream;

    /** Stream for logging output */
    FileWriter logFile;

    /** Create the listener. globalHash0 is an already-created
      Hashtable, and stream0 indicates whether to listen on a Stream
      or Datagram socket. */
    public Listener(Hashtable globalHash0, boolean stream0, FileWriter logFile0) throws SocketException, IOException{

	globalHash=globalHash0;
    
	stream=stream0;

	logFile=logFile0;
	
	if (stream){
	    requestSock = new ServerSocket(Spand.RequestPort);
	} else {
	    /* Listen for Messages on Dgram sock */
	    requestDgramSock = new DatagramSocket(Spand.RequestPort);
	}
      
	this.start();
    }

    /** Start the listener, receiving messages, running their handlers,
      and returning their reply messages. */
      
    public void run() {
	Msg msg = null;
	Msg reply = null;
	SpandHandler handler;
    
	while (true) {
	  /* Receive Message on Listen Port */
	  msg = GetMsg();
	  if (msg!=null){
	    try{
	      /* Try to create the handler that they specified */
	      MsgHandler msgHandler = new MsgHandler(msg, globalHash, 
						     stream, 
						     requestDgramSock, 
						     activeConn,
						     logFile);
	    } catch (Exception e){
	      try{
		logFile.write("+++ ["+new Date().toString()+"] "+msg.sendAddr.toString()+"."+msg.sendPort+" Error: Couldn't process request.."+e.toString()+"\n");
		logFile.flush();
	      } catch (IOException e1){
		System.err.println("Couldn't write to log file");
	      }
	    }
	  }
	} 
    }

    /** Receive a single message from either the stream or datagram socket. */
    private Msg GetMsg(){
	Msg msg = null;

	try {
	  if (stream){
	    /* Accept a Connection from Server Socket */
	    activeConn = requestSock.accept();
	    msg = new Msg(activeConn);
	  } else {
	    msg = new Msg(requestDgramSock);
	  }
	} catch (IOException e){
	  try{
	    logFile.write("+++ Error: Couldn't receive request.."+e.toString()+"\n");
	    logFile.flush();
	  } catch (IOException e1){
	    System.err.println("Couldn't write to log file");
	  }
	}
	return msg;
    }
}




