/*
 * Copyright (c) 1996-1998 Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the SPAND Research
 *      Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/* $Header: /home/stemm/cvs/Classes++/Classes.h,v 1.13 1998/06/10 23:15:40 stemm Exp $ */
#ifndef MS_CLASSES
#define MS_CLASSES

#include <gdbm.h>


/**
 * A List of Objects. The list grows as necessary as objects are added.
 * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.13 $.
 */
template<class T> class List {
 private:
  /** An array of objects */
  T* v;
  
  /* The size of the array of objects */
  int size;
  
 public:
  /** Default constructor */
  List() {v=(T*)malloc(100*sizeof(T)); size=100;}
  
  /** Constructor with a preset size */
  List(int size0) {v=(T*)malloc(size0*sizeof(T)); size=size0;}

  /** Destructor */
  ~List() { free(v);}

  /** The vector operator. If index is > size, then the array is
      resized and the data elements copied. */
  T& operator[](int i)
    {
      T* v1;
      int oldsize;
      
      if (i>=size) {
	oldsize=size;
	size=i*2;
	v1=(T*) malloc(size*sizeof(T));
	memcpy(v1, v, oldsize*sizeof(T));
	free(v);
	v=v1;
      }
      return v[i];
    }
};

/** A Hash from arbitrary object to arbitrary object. The Key and Data
    classes must have Marshal() methods. The marshalled byte string is
    used as the key to the hash. Pointers to the objects are the items
    stored in the hash.
 * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.13 $.
 */
template <class Key, class Data> class Hash {
 public:

  /** Initialize the hash by opening the gdbm file */
  Init(char *filename0) {
    if ((hashDB=gdbm_open(filename0, 0, GDBM_NEWDB, 0777, NULL))==NULL){
      perror("Problem opening hash table");
    }
    
    /* Allocate space for the key and data */
    keySize=dataSize=10000;
    keyMarshalled=(u_char *) malloc(keySize*sizeof(char));
    dataMarshalled=(u_char *) malloc(dataSize*sizeof(char));
  }

  /** Default Constructor */
  Hash() { Init("/tmp/foo.db"); }

  /** Constructor that gives name of gdbm file */
  Hash(char *filename0){
    Init(filename0);
  }

  /** Desctructor, closes gdbm file */
  ~Hash(){
    /* Close database */
    gdbm_close(hashDB);
  }
  
  /** Does lookup in hash, returning pointer to Data object. This
      object should not be freed or deleted. */
  Data *Get(Key &key){
    Data *ptr;
    
    dbKey.dsize=key.Marshal(keyMarshalled);
    dbKey.dptr=(char *) keyMarshalled;
    
    dbVal=gdbm_fetch(hashDB, dbKey);
    
    if (dbVal.dptr==NULL){
      return NULL;
    }
    
    /* If here, then object was found */
    ptr=*((Data **) dbVal.dptr);
    if (dbVal.dptr!=NULL){
      free(dbVal.dptr);
      dbVal.dptr=NULL;
    }
    return ptr;
  }

  /** Put the Key-Data binding in the hash. The Key and Data objects
      are copied into the hash, so the caller is free to delete the
      objects as it sees fit. */
  void Put(Key &key, Data *data){
    /* Resize the marshalled arrays */
    Resize(key.Marshal(NULL));
    Resize(data->Marshal(NULL));
    dbKey.dsize=key.Marshal(keyMarshalled);
    dbKey.dptr=(char *) keyMarshalled;
    
    dbVal.dsize=sizeof(Data *);
    dbVal.dptr=(char *) &data;
    
    if ((rc=gdbm_store(hashDB, dbKey, dbVal, GDBM_REPLACE))==-1){
      perror("Problem with putting into hash");
    }
  }

  /** Delete the binding for the specified Key. */
  void Del(Key &key){
    Data *ptr;
    
    dbKey.dsize=key.Marshal(keyMarshalled);
    dbKey.dptr=(char *) keyMarshalled;
    
    if ((rc=gdbm_delete(hashDB, dbKey))!=0){
      perror("Problem with deleting from hash");
      return;
    }
  }

  /** This is used to start sequential access to the hash and returns
      the first key in the hash. */
  Data *First(){
    Data *ptr;
    
    dbKey=gdbm_firstkey(hashDB);
    if (dbKey.dptr==NULL){
      return NULL;
    }
    
    lastKey=dbKey;
    
    dbVal=gdbm_fetch(hashDB, dbKey);
    ptr=*((Data **) dbVal.dptr);
    if (dbVal.dptr!=NULL){
      free(dbVal.dptr);
      dbVal.dptr=NULL;
    }
    return ptr;
  }

  /** This is used to continue sequential access to the hash */
  Data *Next(){
    Data *ptr;
    
    dbKey=gdbm_nextkey(hashDB, lastKey);
    
    if (dbKey.dptr==NULL){
      return NULL;
    }
    
    if (lastKey.dptr!=NULL){
      free(lastKey.dptr);
      lastKey.dptr=NULL;
    }
    
    lastKey=dbKey;
    
    dbVal=gdbm_fetch(hashDB, dbKey);
    ptr=*((Data **) dbVal.dptr);
    if (dbVal.dptr!=NULL){
      free(dbVal.dptr);
      dbVal.dptr=NULL;
    }
    
    return ptr;
  }
  
 protected:

  int rc;
  datum dbKey, dbVal, lastKey;
  GDBM_FILE hashDB;
  u_char *keyMarshalled;
  int keySize;
  u_char *dataMarshalled;
  int dataSize;

  void Resize(size_t size){
    if (size>keySize){
      free(keyMarshalled);
      keySize=2*size;
      keyMarshalled=(u_char *) malloc(keySize*sizeof(char));
    }
    if (size>dataSize){
      free(dataMarshalled);
      dataSize=2*size;
      dataMarshalled=(u_char *) malloc(dataSize*sizeof(char));		  
    }
  }
  

  };

  /** Like a hash, but the actual objects are stored rather than just
      pointers to them. The caller is responsible for deleting objects
      returned from Get()
      * @author SPAND Group, UC Berkeley.
      * @version $Revision: 1.13 $.
      */
template<class Key, class Data> class PersistentHash: public Hash<Key,Data> {
 public:

  /** Default constructor, just calls parent constructor. */
  PersistentHash(): Hash<Key,Data>() {};

  /** Constructor from file, calls parent constructor. */
  PersistentHash(char *filename0): Hash<Key,Data>(filename0) {};
  
  /** Returns a pointer to a Data Object (which the caller must delete later) */
  Data *Get(Key &key){
    Data *ptr;
    dbKey.dsize=key.Marshal(keyMarshalled);
    dbKey.dptr=(char *) keyMarshalled;
    
    dbVal=gdbm_fetch(hashDB, dbKey);
    
    if (dbVal.dptr==NULL){
      return NULL;
    }
    
    /* If here, then object was found */
    ptr = new Data();
    ptr->Unmarshal((unsigned char *) dbVal.dptr);
    return(ptr);
  }

  /** Adds the key-data binding to the hash. */
  void Put(Key &key, Data *data){
    /* Resize the marshalled arrays */
    Resize(key.Marshal(NULL));
    Resize(data->Marshal(NULL));
    dbKey.dsize=key.Marshal(keyMarshalled);
    dbKey.dptr=(char *) keyMarshalled;
    
    dbVal.dsize=data->Marshal(dataMarshalled);
    dbVal.dptr=(char *) dataMarshalled;
    
    if ((rc=gdbm_store(hashDB, dbKey, dbVal, GDBM_REPLACE))==-1){
      perror("Problem with putting into hash");
    }
  }
  
  /** This is used to start sequential access to the hash and returns
      the first key in the hash. This returns a pointer to a Data
      which the caller must free later. */
  Data *First(){
    Data *ptr;
    
    dbKey=gdbm_firstkey(hashDB);
    if (dbKey.dptr==NULL){
      perror("Problem with getting first from hash");
      return NULL;
    }
    
    lastKey=dbKey;
    
    dbVal=gdbm_fetch(hashDB, dbKey);
    ptr=new Data();
    ptr->Unmarshal((unsigned char *) dbVal.dptr);
    if (dbVal.dptr!=NULL){
      free(dbVal.dptr);
      dbVal.dptr=NULL;
    }
    return ptr;
  }
  
  /** This is used to continue sequential access to the hash. It
      returns a pointer to a Data which the caller must free later. */
  Data *Next(){
    Data *ptr;
    
    dbKey=gdbm_nextkey(hashDB, lastKey);
    
    if (dbKey.dptr==NULL){
      return NULL;
    }
    
    if (lastKey.dptr!=NULL){
      free(lastKey.dptr);
      lastKey.dptr=NULL;
    }
    
    lastKey=dbKey;
    
    dbVal=gdbm_fetch(hashDB, dbKey);
    ptr = new Data();
    ptr->Unmarshal((unsigned char *) dbVal.dptr);
    if (dbVal.dptr!=NULL){
      free(dbVal.dptr);
      dbVal.dptr=NULL;
    }
    
    return ptr;
  }
  
};

/** This Object represents a Collection of Objects. Add, Delete, and
    sequential access are supported.
 * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.13 $.
 */
template<class T> class Collection {
 public:

  /** Default Constructor */
  Collection() {
    nElem=0;
  }

  /** Add the element to the collection */
  void Add(T *add) {
    l[nElem]=*add;
    nElem++;
  }

  /** Remove the element from the collection */
  void Remove(T* del) {
    int i;
    
    for(i=0; i<nElem; i++){
      if (l[i]==*del) {
	nElem--;
	l[i]=l[nElem];
      }
    }
  }

  /** This is used to start sequential access to the collection and
      returns the first key in the hash. This returns a pointer to a
      Data which the caller must not modify or delete. */
  T *First() {
    count=0;
    if (nElem==0){
      return NULL;
    } else {
      return &(l[count]);
    }
  }

  /** This is used to continue sequential access to the hash. It
      returns a pointer to a Data which the caller must not modify or
      delete. */
  T *Next() {
    count++;
    if (count<nElem) {
      return &(l[count]);
    } else {
      return NULL;
    }
  }
  
  /** The number of elements in the collection */
  int nElem;
  
 protected:
  /** The collection is internally represented as a list */
  List<T> l;

  /** A counter used as a placeholder for sequential access. */
  int count;

};

/** This extends a Collection by adding the ability to Marshal the
    collection into a sequence of bytes.
 * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.13 $.
 */
template<class T> class MarshalCollection: public Collection<T> {
 public:
  /** Default Constructor, just calls parent constructor */
  MarshalCollection():Collection<T> () {};

  /** Constructor from sequence of bytes, calls Unmarshal method. */
  MarshalCollection(u_char *wireData) : Collection<T> () {
    Unmarshal(wireData);
  }
  
  
  /** Turns Collection into sequence of bytes by calling Marshal on
      each member of collection. Also saves number of items for
      Unmarshaling. */
  u_long Marshal(u_char *wireData){
    u_long i;
    struct WireFormat *m = (struct WireFormat *) wireData;
    u_char *p = (u_char *) m+sizeof(struct WireFormat);
    u_long size;
    
    if (wireData==NULL){
      size=0;
      for(i=0; i<nElem; i++){
	size+=l[i].Marshal(NULL);
      }
      return size;
    }
    
    m->count=count;
    m->nElem=nElem;
    for(i=0; i<nElem; i++){
      p+=l[i].Marshal(p);
    }
    
    return (u_long)p-(u_long) wireData;
  }
  
  /** Unpackages collection from sequence of bytes by calling
      Unmarshal count number of times. */
  u_long Unmarshal(u_char *wireData){
    u_long i;
    struct WireFormat *m=(struct WireFormat *) wireData;
    u_char *p = (u_char *) m+sizeof(struct WireFormat);
    u_long size;
    
    count=m->count;
    nElem=m->nElem;
    for(i=0; i<nElem; i++){
      p+=l[i].Unmarshal(p);
    }
    
    return (u_long)p-(u_long) wireData;
  }
 protected:
  /** Wire Format of Collection. Multi-byte fields are kept in network
      order. */
  struct WireFormat {
    int count;
    int nElem;
  };

};
#endif


