/*
 * Copyright (c) 1996-1998 Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the SPAND Research
 *      Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/* $Header: /home/stemm/cvs/Socket++/Socket.h,v 1.7 1998/06/08 15:14:55 ss Exp $ */
#ifndef SS_SOCKET
#define SS_SOCKET

#include "SocketTypes.h"
#include "Utility.h"
#include "Channel.h"
#include <Classes.h>

/** Abstract class that all the Socket classes derive from the class. It
  defines the basic interface that all sockets should be able to
  support. */
class Socket : public Channel {
public:
	
	Socket() {};
	/** creates a new control or normal communication channel */
	Socket(int isCtrl);
	virtual ~Socket();
	
	/* standard BSD socket routines */
	
	/** connect socket to remote socket */
	Sock_Status Connect(char *hostname, Port port);
	/** listen for connections on port */
	Sock_Status Listen(Port port);
	/** accept a connection on a port 
	  return val: socket connected to hostname */
	virtual Socket* Accept() = 0;
	/** accept a connection on a port -- socket cannot be accepted again*/
	Sock_Status AcceptOW();
	/** close socket and connection and then free associated storage */
	Sock_Status Close();
	
	/* set of options on sockets */
	/** prevent routines from trying to resolve any addresses to
	  names. Useful if name server is down */
	inline void SetResolve(int value) {resolve_ = value;};
	/** set max transmission rate */
	inline void SetXmitRate(int value) {xmitRate_ = value;};
	/** Whether to listen on and use the control channel 
	  associated with data channel */
	inline void SetUseCtrl(int value) {useCtrl_ = value;};
	/** Ensure that the library does not send any extra
	  data that is used by the library itself. Prevents the
	  creation of a control connection, etc. This makes it
	  compatible with other programs not using this library. */
	inline void SetCompatible(int value) {compatible_ = value;};
	/** value to set flags for sendto calls */
	inline void SetSendFlag(int value) {lFlags_ = value;};
	/** set MSG_OOB flag for sendto calls */
	inline void SetOOB(int value) {if (!value) lFlags_|=MSG_OOB;};
	/** will cause binding of socket to passed addr. Must be
	  called before Listen */
	inline void SetAcceptFlag(int addr) {acceptFlag_=1;
	lSin_.sin_addr.s_addr=
		(u_long) htonl((u_long) addr);};
	/** set the Socket to blocking or non-blocking */
	inline int SetBlock(int value) {block_ = value ? SOCK_BLOCK:SOCK_NON_BLOCK;
	return(SetFlag(O_NDELAY, value));};
#ifdef IP_TOS	 
	/** set the TOS on outgoing packets */
	inline int SetTOS(int value) {return(SetSockOpt(IP_TOS,value,"ip"));};
#endif
	/** set the timeout for all calls */
	int SetTimeout(Usec value);
	/** set the socket buffer size */
	int SetBufferSize(int bufSize);
	/** set socket flags using fcntl */
	int SetFlag(int flag, int value);
	/** set socket options directly */
	int SetSockOpt(int flag, int value, char *layer);
	
	/* get value of options on sockets */
	
	/** get setting of resolve */
	inline int GetResolve() {return resolve_;};
	/** get setting of xmit rate */
	inline int GetXmitRate() {return xmitRate_;};
	/** get setting of using a control socket */
	inline int GetUseCtrl() {return useCtrl_;};
	/** get setting of compatibility setting */
	inline int GetCompatible() {return compatible_;};
	/** get setting of send flags */
	inline int GetSendFlag() {return lFlags_;};
	/** get OOB setting of send flags */
	inline int GetOOB() {return lFlags_ & MSG_OOB;};
	/** get setting of local binding of socket */
	inline int GetAcceptFlag() {if (acceptFlag_)
		return lSin_.sin_addr.s_addr;
	else return 0;};
	/** get setting of blocking status */
	inline int GetBlock() {return(GetFlag(O_NDELAY));};
#ifdef IP_TOS	 
	/** get setting of output TOS */
	inline int GetTOS() {return(GetSockOpt(IP_TOS,"ip"));};
#endif
	/** get setting of timeout */
	inline int GetTimeout() {return(Sock_TvalToUtime(toutPtr_));};
	/** get setting of socket buffer size */
	inline int GetBufferSize() {return lRecvBufSize_;};
	/** get file descriptor used by socket */
	inline GetFD() {return rfd_;};
	/** get local port number */
	inline GetLocalPort() {return localPort_;};
	/** get remote port number */
	inline GetRemotePort() {return remotePort_;};
	/** get port number of TCP control connection */
	inline GetTCPCtrlPort() { if (tcpCtrlChannel_ == NULL) return SOCK_FAILURE;
	else return tcpCtrlChannel_->localPort_;};
	/** get port number of UDP control connection */
	inline GetUDPCtrlPort() { if (udpCtrlChannel_ == NULL) return SOCK_FAILURE;
	else return udpCtrlChannel_->localPort_;};
	/** get socket flags using fcntl */
	int GetFlag(int flag);
	/** get type of socket (UDP,TCP,etc) */
	int GetType() {return type_;};
	/** get socket options set */
	int GetSockOpt(int flag, char *layer);
	/** get name of remote host */
	char *GetRemoteName();
	/** get name of local host */
	char *GetLocalName();
	/** get addr of remote host */
	Address *GetRemoteAddr();
	/** get addr of local host */
	Address *GetLocalAddr();
	/** get addr of last remote host to send data */
	Address *GetLastSenderAddr();

	/** XXX Not Implemented */
	virtual Marshal(u_char *wireData) {};
	/** XXX Not Implemented */
	virtual Unmarshal(u_char *wireData) {};

protected:
	int RecvStream(char *buffer, int length);
	int RecvStreamFull(char *buffer, int length);
	int RecvDgram(char *buffer, int length);
	int RecvDgramFull(char *buffer, int length);
	int SendStream(char *buffer, int length);
	int SendDgram(char *buffer, int length);
	int WaitForPktRecv();
	int WaitForPktSend(int);
	Sock_Status SetSocketInfo();
	int SetRemoteByName(char *hostname, Port port);
	int SetRemoteByAddr(Address *rSinPtr);
	int SendHello();
	int RecvHello();
	
	virtual Sock_Status InitSocket() = 0;
	virtual inline int accept_(int fd, struct sockaddr *sin, int *len) 
#ifdef AIX4
		{return(accept(fd, sin, (long unsigned int *) len));};
#elif (defined LINUX)
	{return(accept(fd, sin, (unsigned int *) len));};
#else
	{return(accept(fd, sin, len));};
#endif
	virtual inline int connect_(int fd, struct sockaddr *sin, int len)
		{return(connect(fd, sin, len));};
	virtual inline int listen_(int fd, int count)
		{return(listen(fd, count));};
	virtual inline int bind_(int fd, struct sockaddr *sin, int len)
		{return(bind(fd, sin,len));};
	virtual inline int getsockname_(int fd, struct sockaddr *sin, int *len)
#ifdef AIX4
		{return(getsockname(fd, sin, (long unsigned int *)len));};
#elif (defined LINUX)
	{return(getsockname(fd, sin, (unsigned int *) len));};	
#else
	{return(getsockname(fd, sin, len));};
#endif
	virtual inline int select_(int width, fd_set *rfds, fd_set *wfds, 
				   fd_set *efds, struct timeval *tout)
		{return(select(width, rfds, wfds, efds, tout));};

	void CheckCtrl();
	int RecvCtrl(SockCtrlPkt *ctrlPktPtr);
	int SendCtrl(RemoteHost *rhostPtr, SockCtrlPkt *ctrlPktPtr);
	int CtrlSendFarewell(RemoteHost *rhostPtr);
	void ProcessCtrlPkt(SockCtrlPkt *);
	
	/***************** basic info *****************/
	int rfd_;
	int sfd_;
	int state_;
	int side_;
	int type_;
	PROTOCOLS protocol_;
	
	/***************** local info *****************/
	char localName_[SOCK_MAX_HOST_NAME_LEN];
	char localIPAddr_[SOCK_MAX_IP_LEN];
	Port localPort_;
	Address lSin_;
	int lRecvBufSize_;
	int lXmitBufSize_;
	
	/**************** how to send stuff out ******************/
	int acceptFlag_;		/* 1: bind to fixed address; 0: INADDR_ANY */
	int lFlags_;
	char ttl_;
	timev timeout_, *toutPtr_;
	/*    BerState berState_; */
	int xmitRate_;
	struct timeval lastPktTime_;
	int lastPktSize_;
	int broadcast_;
	int isCtrl_;
	int useCtrl_;
	int block_;
	int compatible_;
	int resolve_;
	
	/***************** Control channel stuff *****************/
	struct Socket *udpCtrlChannel_;
	struct Socket *tcpCtrlChannel_;
	int numActiveRHost_;
	RemoteHost *rHost_;
	Port remotePort_;
	Address rSin_;
	Address lastRSin_;
};



/*****************************************************************
  remote ctrl routines
  *************************************************************************/

/* some useful ctrl functions that affect the remote side of connection
   you must set_ctrl on both sides before using these
   
   all Sockets have an global ctrl channel for creating new connections
   (for multicast) and processing ctrl info from machines not "connected" to
   this socket. All machines that are "connected" to this socket have separate
   ctrl channels.
   This is global ctrl channel is a udp socket, the port
   number is obtained by using Sock_GetLocalCtrlPort().
   */

/* if you would like to ctrl the connecttion between a set of hosts from a
   outside (third-party) host you should create a ctrl-only channel with
   Sock_CtrlNewCtrl()
   */
#if 0
extern Socket           *Sock_CtrlNewCtrl(Socket *ctrlSdPtr, char *hostname,
					  Port port, int protocol);

/* set transmission rate of remote side */
extern Sock_Status      Sock_CtrlSetRate(Socket *ctrlSdPtr, int rate);

/* force remote side to close */
extern Sock_Status      Sock_CtrlClose(Socket *ctrlSdPtr, char *localName);
#endif

Socket *Select(char *readWriteExec, int usec, Collection<Socket *> &socks);

#include "Tcp.h"
#include "Udp.h"
#include "UdpMulti.h"
#include "UdpReliable.h"
#include "Utility.h"
#include "Raw.h"
#endif

