/*
 * Copyright (c) 1996-1998 Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the SPAND Research
 *      Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/* $Header: /home/stemm/cvs/Socket++/UdpReliable.cc,v 1.4 1998/05/28 20:15:30 stemm Exp $ */
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <assert.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <math.h>
#include "porting.h"
#include "dbgprintlib.h"
#include "Socket.h"
#include "Udp.h"
#include "UdpMulti.h"


UdpReliableSocket::UdpReliableSocket() : UdpSocket(SOCK_NOT_CTRL)
{
    if (InitSocket() != SOCK_SUCCESS) {
	state_ = SOCK_STATE_CLOSED;
	return;
    }
}

UdpReliableSocket::UdpReliableSocket(int isCtrl = SOCK_NOT_CTRL) : UdpSocket(isCtrl)
{
    if (InitSocket() != SOCK_SUCCESS) {
	state_ = SOCK_STATE_CLOSED;
	return;
    }
}

UdpReliableSocket::~UdpReliableSocket() 
{
    int i;

    /* free the cache and all that state */
    for (i = 0; i < SOCK_MAXWIND; i++) {
	free(cache_->pkts[i]);
	free(cache_->rcvPkts[i]);
    }
    free(cache_);
}
    
Sock_Status
UdpReliableSocket::InitSocket()
{
    Sock_Status status;
    int i;

    status = UdpSocket::InitSocket();
    protocol_ = SOCK_UDP_RELIABLE;
    delack_ = 1; /* default number of packets before sending ack */
    cache_ = (SockCache *) Sock_Malloc(sizeof(SockCache));
    cache_->lastSent = cache_->lastRead = cache_->lastAck = 
	cache_->ackBase = -1;
    cache_->ackMask = cache_->delack = 0;
    cache_->srtt = SOCK_INITRTT;
	  
    for (i = 0; i < SOCK_MAXWIND; i++) {
	cache_->pkts[i] = (SockPkt *) Sock_Malloc(sizeof(SockPkt));
	cache_->pkts[i]->localHdr.numRxmit = 0;
	cache_->pkts[i]->validFlag = SOCK_AVAIL;
	cache_->rcvPkts[i] = (SockPkt *) Sock_Malloc(sizeof(SockPkt));
	cache_->rcvPkts[i]->validFlag = SOCK_AVAIL;
    }
    return(SOCK_SUCCESS);
}

Sock_Status
UdpReliableSocket::Connect(char *hostname, Port port)
{
    Socket::Connect(hostname, port);

    if (compatible_ == SOCK_NOT_COMPATIBLE && isCtrl_ != SOCK_CTRL) {
	int status;

	if ((status = SendHello()) != SOCK_SUCCESS)
	    return(status);
	if ((status = RecvHello()) != SOCK_SUCCESS)
	    return(status);
    }
    if (connect_(sfd_, (struct sockaddr *)&rSin_, sizeof(rSin_)) == -1) {
	debug_printf(110, "Sock_connect: Error connecting");
	debug_printf(110, " errno %d\n", errno);
	return(SOCK_FAILURE);
    }
    numActiveRHost_ = 1;

    SetSocketInfo();
    state_ = SOCK_STATE_CONNECTED;
    return(SOCK_SUCCESS);
}

UdpReliableSocket *
UdpReliableSocket::Accept()
{
    int temp;
    UdpReliableSocket *cdPtr;

    cdPtr = new UdpReliableSocket(isCtrl_);
    
    *cdPtr = *this;
    temp = cdPtr->AcceptOW();
    if (temp != SOCK_SUCCESS) {
	delete cdPtr;
	cdPtr = NULL;
    } else {
	rfd_ = -1;
	sfd_ = -1;
    }
    return(cdPtr);
}

Sock_Status
UdpReliableSocket::Recv(char *buffer, int length)
{
    Sock_Status status;

    debug_printf(112, "UdpRel_Recv %d\n", length);
    status = RecvReliableDgram(buffer, length, NULL);
    return status;
}

Sock_Status
UdpReliableSocket::RecvFull(char *buffer, int length)
{
    Sock_Status status;

    debug_printf(112, "UdpRel_RecvFull %d\n", length);
    status = RecvReliableDgram(buffer, length, NULL);
    return status;
}

/* 
 * Send datagrams reliably. The data to be sent is in char *buffer and has 
 * of 'length' bytes.
 */
Sock_Status
UdpReliableSocket::Send(char *buffer, int length)
{
    int index;
    SockPkt *p;
    timev *timeout;
     Sock_Status status;

    debug_printf(112, "UdpRelSock_Send %d\n", length);
    WaitForPktSend(length);

    p = MakePkt(buffer, ++cache_->lastSent, 0, 0, length);
    index = cache_->lastSent % SOCK_MAXWIND;
    if (InsertPkt(p) == SOCK_FAILURE)
	while (cache_->pkts[index]->validFlag == SOCK_UNAVAIL) {
	    /* 
	     * Can't send because this space is filled up, so try receiving 
	     * instead and hope that we get an ack soon to free up the slot.
	     * XXXXX This does not look correct --- Srini
	     */
	    timeout = Sock_UtimeToTval(2*cache_->srtt);
	    debug_printf(155, "rtt %d, sec %d, usec %d\n", cache_->srtt, 
		    timeout->tv_sec, timeout->tv_usec);
	    if (RecvReliableDgram(buffer, sizeof(SockSendHdr),
				       timeout) == SOCK_TIMEOUT)
		RexmtPkt(cache_->pkts[index]);
	    if (InsertPkt(p) != SOCK_FAILURE)
		break;
	}
    status = SendPkt(p); /* send only if successfully buffered */
    if (status < 0) 
	lastPktSize_ = 0;
    return (status);
}

Sock_Status 
UdpReliableSocket::RecvReliableDgram(char *buffer, unsigned int length, 
				     timev *timeout)
{
     int nread;
     SockPkt *pkt = cache_->rcvPkts[(cache_->lastRead + 1) % SOCK_MAXWIND];
     Sock_Status status;

    if ((status = WaitForPktRecv()) != SOCK_SUCCESS)
	return status;

     if (pkt->validFlag == SOCK_UNAVAIL) {
	  debug_printf(155, "read %d from rcvbuf\n", pkt->sendHdr.seq);
	  ++cache_->lastRead;
	  pkt->validFlag = SOCK_AVAIL;		    
	  memcpy(buffer, pkt->buf, 
		 (size_t) (nread = SOCK_MIN(length, pkt->sendHdr.size)));
	  if (length != pkt->sendHdr.size)
	       debug_printf(110, "incomplete recv: length %d, read %d\n");
	  if (pkt->sendHdr.pktFlags & SOCK_ACK)
	       ProcessAck(pkt->sendHdr.ack, pkt->sendHdr.ackBits);
	  free((void *) pkt->buf);
	  return nread;
     }
     return ReadPkt(buffer, length, timeout);
}

int 
UdpReliableSocket::ReadPkt(char *buffer, int length, timev *timeout)
{
     struct iovec iov[2];
     unsigned int hdrSize = sizeof(SockSendHdr), totalLen;
     char *hdrBuf = (char *) Sock_Malloc(hdrSize);
     SockPkt *pkt = cache_->rcvPkts[(cache_->lastRead + 1) % SOCK_MAXWIND];
     SockSendHdr *recvHdr;


     iov[0].iov_base = (caddr_t) hdrBuf;
     iov[0].iov_len = hdrSize;
     iov[1].iov_base = (caddr_t) buffer;
     iov[1].iov_len = length;
	  
     while (1) {
	  if ((totalLen = readv(rfd_, &iov[0], 2)) < hdrSize)
	       return SOCK_FAILURE;
	  recvHdr = (SockSendHdr *) hdrBuf;
	  recvHdr->seq = ntohl(recvHdr->seq);
	  recvHdr->ack = ntohl(recvHdr->ack);
	  recvHdr->ackBits = ntohl(recvHdr->ackBits);
	  recvHdr->size = ntohl(recvHdr->size);
	  
	  if (totalLen - hdrSize != recvHdr->size) {
	      /* error in readv */
	      debug_printf(119, "ReadPkt: total %d, hdrsize %d, length %d\n",
		      totalLen, hdrSize, length);
	      return SOCK_FAILURE;
	  }

	  if (recvHdr->pktFlags & SOCK_FIN) {
	      Close();
	      free((void *) hdrBuf);
	      return 0;	/* ??? */
	  }

	  if (recvHdr->pktFlags & SOCK_ACK) {
	      ProcessAck(recvHdr->ack, recvHdr->ackBits);
	      if (recvHdr->size == 0) /* pure ack */
		  break;	
	  }
	  
	  if (recvHdr->seq == cache_->lastRead + 1) { /* common case */
	      ++cache_->lastRead;
	      do {
		  /*cache_->ackMask >> 1;*/ /* doesn't work under Solaris */
		  cache_->ackMask /= 2;
		  ++cache_->ackBase;
	      } while (cache_->ackMask % 2);
	      if (SendAck(SOCK_DELACK) == SOCK_FAILURE)
		  debug_printf(110, "Error sending ack\n");
	      break;

	  } else if (recvHdr->seq > cache_->lastRead) {
	      int index = recvHdr->seq % SOCK_MAXWIND;
	      int diff;

	      debug_printf(155, "out of seq %d, lastread %d\n", recvHdr->seq, 
		      cache_->lastRead);
	      pkt = cache_->rcvPkts[index];
	      if (pkt->validFlag & SOCK_UNAVAIL) {
		  /* drop the packet, don't generate ack, read again.
		     actually, this couldn't possibly have happened. */
		  debug_printf(110, "readpkt: inconsistent state\n");
		  continue;
	      }
	      /* insert pkt in rcvPkts cache (recvHdr->size > 0) */
	      pkt->validFlag = SOCK_UNAVAIL;
	      pkt->buf = (char *) Sock_Malloc(recvHdr->size); /* XXX */
	      memcpy(pkt->buf, buffer, recvHdr->size);
	      pkt->sendHdr.seq = recvHdr->seq;
	      pkt->sendHdr.ack = recvHdr->ack;
	      pkt->sendHdr.size = recvHdr->size;
	       
	      diff = recvHdr->seq - cache_->ackBase;
	      cache_->ackMask |= (1 << (diff - 1));
	      /* send an ack with ack field set to recvHdr->seq */
	      if (SendAck(SOCK_NODELACK) == SOCK_FAILURE)
		  debug_printf(110, "Error sending ack\n");
	  } else
	      /* sender retransmitted packet, maybe ack got lost */
	      SendAck(SOCK_NODELACK);
     }
     free((void *) hdrBuf);
     return(recvHdr->size);
}


int 
UdpReliableSocket::SendPkt(SockPkt *pkt)
{
    struct iovec iov[2];
    unsigned int size;
    int writeErr;

    iov[0].iov_base = (char *)&(pkt->sendHdr);
    iov[0].iov_len = sizeof(SockSendHdr);
    iov[1].iov_base = pkt->buf;
    iov[1].iov_len = pkt->localHdr.size;

    size = pkt->localHdr.size;
    gettimeofday(&(pkt->localHdr.sndTime), NULL);
     
    writeErr = writev(sfd_, &iov[0], 2);
    if (writeErr != (int) (sizeof(SockSendHdr) + size)) {
	if (writeErr < 0)
	    printf("SendPkt: writeErr %d, errno %d\n", writeErr, errno);
	debug_printf(110, "Sock_SendPkt: error in writev");
	if (errno == EPIPE) {
	    state_ = SOCK_STATE_CLOSED;
	    return(SOCK_FD_CLOSED);
	}
	return(SOCK_FAILURE);
    }
    debug_printf(155, "sent %d\n", pkt->localHdr.seq);
    return(size);
}

int 
UdpReliableSocket::RexmtPkt(SockPkt *pkt)
{
    debug_printf(140, "rexmit %d\n", pkt->localHdr.seq);
    ++pkt->localHdr.numRxmit;
    return SendPkt(pkt);
}


int 
UdpReliableSocket::SendAck(int delack)
{
    SockSendHdr sendHdr;
    char *buffer = (char *) &sendHdr;
    int status;

    if (delack == SOCK_DELACK)	/* generate a delayed ack */
	if (++cache_->delack != delack_)
	    return SOCK_SUCCESS;

    cache_->delack = 0; /* reset number of delayed ack waits */
    if (cache_->ackMask != 0)
	debug_printf(150,"sent ack %d mask %x\n", cache_->ackBase, cache_->ackMask);
    else
	debug_printf(155,"sent ack %d mask %x\n", cache_->ackBase, cache_->ackMask);
    sendHdr.pktFlags = SOCK_ACK;
    sendHdr.seq = htonl(++cache_->lastSent);
    sendHdr.ack = htonl(cache_->ackBase); /* cumulative ack */
    sendHdr.ackBits = htonl(cache_->ackMask); /* bit pattern of holes */
    sendHdr.size = 0;
     
    if ((status = SendDgram(buffer, sizeof(SockSendHdr))) == SOCK_FAILURE)
	debug_printf(110, "error in sending ack, status %d\n", status);
    return status;
    /* Sock_SendPkt needs a connect done from the receiver as well
       return Sock_SendPkt(sdPtr, pkt);
       */
}
     
inline void
UdpReliableSocket::FreePkts(int index, SockSeq pktSeq, SockSeq baseAck)
{
    SockSeq seq;
    int i = index;
    /* cleans whole cache each time */
    do {
	seq = cache_->pkts[i]->localHdr.seq;
	if (cache_->pkts[i]->validFlag == SOCK_UNAVAIL &&
	    (seq <= baseAck || seq == pktSeq)) {
	    debug_printf(155, "freeing %d at %d\n", seq, index);
	    cache_->pkts[i]->validFlag = SOCK_AVAIL;
	    free(cache_->pkts[i]->buf);
	}
	i = (i + 1) % SOCK_MAXWIND;
    } while (i != index);
}

void
UdpReliableSocket::ComputeRtt(SockSeq baseAck)
{
    int index = baseAck % SOCK_MAXWIND;
    SockPkt *pkt = cache_->pkts[index];
    int rtt;			/* rtt estimate for current packet */
    timev curTime;

    if (pkt->validFlag == SOCK_UNAVAIL && pkt->localHdr.seq == baseAck) {
	gettimeofday(&curTime, NULL);
	rtt = Sock_TimeDiff(curTime, pkt->localHdr.sndTime);
	/* XXX optimize this later */
	cache_->srtt = (int) (SOCK_RTT_MULTIPLIER * rtt + 
	    (1 - SOCK_RTT_MULTIPLIER) * cache_->srtt);
    }
}


int
UdpReliableSocket::ProcessAck(SockSeq baseAck, u_int ackMask)
{
    int index = baseAck % SOCK_MAXWIND, forceRtx = 0, rtxIndex;
    SockSeq lastAck, pktSeq = baseAck;

    if (ackMask != 0)
	debug_printf(150, "ackbase %d, mask %x\n", baseAck, ackMask);
    else
	debug_printf(155, "ackbase %d, mask %x\n", baseAck, ackMask);
    if (ackMask == 0) {
	if (cache_->lastAck < baseAck)
	    ComputeRtt(baseAck);
	cache_->lastAck = baseAck;
	FreePkts(index, pktSeq, baseAck);
	return SOCK_SUCCESS;
    }

    /* ackMask != 0; guaranteed that there is a "hole" */
    lastAck = baseAck;
    rtxIndex = (baseAck + 1) % SOCK_MAXWIND;
    if (ackMask & SOCK_FULLMASK) {
	debug_printf(155, "full mask\n");
	/* current window xmitted, so can rexmit even if done once */
	forceRtx = 1;
    }
    /* clean cache of packets <= baseAck */
    FreePkts(index, baseAck, baseAck); 
    /* now clean cache of higher number packets using ackmask */
    while (ackMask) {
	SockPkt *pkt = cache_->pkts[index];
	index = (index + 1) % SOCK_MAXWIND;
	++pktSeq;
	if (ackMask % 2) 	/* packet has reached receiver */
	    if (pkt->validFlag == SOCK_UNAVAIL && 
		pkt->localHdr.seq == pktSeq) {
		debug_printf(150, "free %d at %d\n", pkt->localHdr.seq, index);
		pkt->validFlag = SOCK_AVAIL;
		free(pkt->buf);
	    }
	++lastAck;
	ackMask /= 2;		/* bit shift breaks under Solaris? */
    }     
     
    if (cache_->lastAck < lastAck)
	cache_->lastAck = lastAck;

    /* 
     * Retransmit packet if a 'hole' is present in the ack sequence, i.e, 
     * if there is an unack'd packet and a later one has been ack'd.
     */
    if (cache_->pkts[rtxIndex]->validFlag == SOCK_UNAVAIL)
	if (cache_->pkts[rtxIndex]->localHdr.numRxmit == 0 || forceRtx) {
	    debug_printf(155, "rtx pkt %d at %d\n", 
		    cache_->pkts[rtxIndex]->localHdr.seq, rtxIndex);
	    RexmtPkt(cache_->pkts[rtxIndex]);
	}
    return SOCK_SUCCESS;
}


/* make a packet to put in the socket cache */
SockPkt *
UdpReliableSocket::MakePkt(char *buffer, SockSeq seq, SockSeq ack, 
			   u_int ackBits, size_t length)
{
    SockPkt *pkt = (SockPkt *) malloc(sizeof(SockPkt));

    pkt->localHdr.seq = seq;
    pkt->localHdr.ack = ack;
    pkt->localHdr.size = length;
    pkt->localHdr.numRxmit = 0;
    gettimeofday(&(pkt->localHdr.sndTime), NULL);

    pkt->sendHdr.seq = htonl(seq);
    pkt->sendHdr.ack = htonl(ack);
    pkt->sendHdr.ackBits = htonl(ackBits);
    pkt->sendHdr.size = htonl(length);
    pkt->sendHdr.pktFlags = 0;

    pkt->buf = (char *) Sock_Malloc(length);
    memcpy(pkt->buf, buffer, length);
    return pkt;
}

/* insert packet into cache and keep it there until we get an ack for it */
int
UdpReliableSocket::InsertPkt(SockPkt *pkt)
{
    int index = cache_->lastSent % SOCK_MAXWIND;

    if (cache_->pkts[index]->validFlag == SOCK_AVAIL) {
	/* this place is available in the cache, so insert packet here */
	pkt->validFlag = SOCK_UNAVAIL;
	cache_->pkts[index] = pkt;
    } else			/* return -1 to let caller block */
	return SOCK_FAILURE;
    return SOCK_SUCCESS;
}
