/*
 * Copyright (c) 1996-1998 Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the SPAND Research
 *      Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/* $Header: /home/stemm/cvs/SPAND/src/libspand/Spand.h,v 1.19 1998/06/17 22:18:48 stemm Exp $ */
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <math.h>
#include <strings.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <porting.h>
#include <Socket.h>
#include <Classes.h>

#define REQUEST_PORT (3000)
#define MAX_MSG_SIZE (1500)

/** The Spand library defines the objects that reprsent Spand Messages
as well as helper objects that send messages to the Performance
Server. The C++ implementation of the Spand library is a subset of the
complete Java implementation, and only contains enough functionality
to implement Spand clients (and not servers). In particular, the C++
implementation does not contain the Listener objects that implement
the core Performance Server functionality. Also missing is the
implementation of the GetPerf, GetSumm, and GetRaw agents.
 * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.19 $.
 */
class Spand {
};

/** A SpandObject is (roughly) the same as an Object class in
   Java. All objects inherit from SpandObject, and all SpandObjects
   must provide a Print() method.
 * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.19 $.
 */
class SpandObject: public Spand{
 public:
  /** Return a String representation of this object. */
  virtual char * Print() = 0;

 protected:
  char buf[MAX_MSG_SIZE];
};

/** A SendableSpandObject is an object that can be packaged up and
   sent over a network. It defines Marshal and Unmarshal methods.
 * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.19 $.
 */
class SendableSpandObject: public SpandObject{
public:

  /** Package this object into a sequence of bytes. The number of
      bytes written is returned. If wireData is NULL, then no bytes
      are actually written, but the size of the marshalled object is
      returned. */
  virtual u_long Marshal(u_char *wireData) = 0;

  /** Unpackage this object from a sequence of bytes. The number of
      bytes read is returned. */
  virtual u_long Unmarshal(u_char *wireData) = 0;

  /** Compare two objects. Two objects are considered equal if their
      Marshalled formats are the same size and equal. */
  int operator==(SendableSpandObject &b);
};

/** An Internet Address Object.
  * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.19 $.
 */
class Addr: public SendableSpandObject {
public:

  /** Default Constructor */
  Addr();
  
  /** Constuctor from in_addr structure. */
  Addr(struct in_addr addr0);

  /** Constructor from hostname string. */
  Addr(char *addr);

  /** Package this object into a sequence of bytes. The number of
      bytes written is returned. If wireData is NULL, then no bytes
      are actually written, but the size of the marshalled object is
      returned. */
  u_long Marshal(u_char *wireData);

  /** Unpackage this object from a sequence of bytes. The number of
      bytes read is returned. */
  u_long Unmarshal(u_char *wireData);

  /** Return a String representation of this object. */
  char * Print();
  
  /** The address is internally represented as a in_addr structure */
  struct in_addr addr;

private:

  /** Initialize the Addr object */
  Init(struct in_addr addr0);

  /** The WireFormat of the Addr object */
  struct WireFormat {
    struct in_addr addr;
  };

};

/** An Internet Port Object
  * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.19 $.
 */
class ReportPort: public SendableSpandObject {
public:

  /** Default Constructor */
  ReportPort();

  /** Constructor from u_short port */
  ReportPort(u_short port0);

  /** Package this object into a sequence of bytes. The number of
      bytes written is returned. If wireData is NULL, then no bytes
      are actually written, but the size of the marshalled object is
      returned. */
  u_long Marshal(u_char *wireData);

  /** Unpackage this object from a sequence of bytes. The number of
      bytes read is returned. */
  u_long Unmarshal(u_char *wireData);

  /** Return a String representation of this object. */
  char * Print();
  
  /** The port is internally represented as a u_short */
  u_short port;

private:

  /** Initialize the ReportPort object */
  Init(u_short port0);

  /** The WireFormat of the ReportPort object. The port is expanded to
      an u_long when Marshalled. */
  struct WireFormat {
    u_long port;
  };

};

/** Address/Port Pair
  * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.19 $.
 */
class AddrPort: public SendableSpandObject {
public:

  /** Package this object into a sequence of bytes. The number of
      bytes written is returned. If wireData is NULL, then no bytes
      are actually written, but the size of the marshalled object is
      returned. */
  u_long Marshal(u_char *wireData);

  /** Unpackage this object from a sequence of bytes. The number of
      bytes read is returned. */
  u_long Unmarshal(u_char *wireData);

  /** Return a String representation of this object. */
  char * Print();
  
  /** The Underlying Address object */
  Addr addr;

  /** The underlying port object */
  ReportPort port;

};

/** A Flow is identified by source addr/port and dest addr/port.
  * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.19 $.
 */
class FlowID: public SendableSpandObject {
public:

  /** Package this object into a sequence of bytes. The number of
      bytes written is returned. If wireData is NULL, then no bytes
      are actually written, but the size of the marshalled object is
      returned. */
  u_long Marshal(u_char *wireData);

  /** Unpackage this object from a sequence of bytes. The number of
      bytes read is returned. */
  u_long Unmarshal(u_char *wireData);

  /** Return a String representation of this object. */
  char * Print();

  /** The Source addr/port. */
  AddrPort src;

  /** The Dest addr/port. */
  AddrPort dst;

};

typedef struct _FlowIDStruct {
  struct in_addr saddr;
  u_long sport;
  struct in_addr daddr;
  u_long dport;
} FlowIDStruct;

/** The default performance report. Contains bandwidth and rtt estimates.
  * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.19 $.
 */
class ReportPayload : public SendableSpandObject{

public:
  /** Default constructor */
  ReportPayload();

  /** Constructor from FlowID object */
  ReportPayload(FlowID id0);

  /** Initializes Fields */
  Init();

  /** Package this object into a sequence of bytes. The number of
      bytes written is returned. If wireData is NULL, then no bytes
      are actually written, but the size of the marshalled object is
      returned. */
  u_long Marshal(u_char *wireData);

  /** Unpackage this object from a sequence of bytes. The number of
      bytes read is returned. */
  u_long Unmarshal(u_char *wireData);

  /** Return a String representation of this object. */
  char * Print();
  
  /** The Flow Identifier for this report */ 
  FlowID id;

  /** The time the report was made in seconds since the epoch */
  u_long mostSigNTP;

  /** The time the report was made in microseconds sice the epoch */
  u_long leastSigNTP;

  /** The number of bytes in this report */
  u_long sampLen;

  /** The duration of the sample in milliseconds */
  u_long sampDur;

  /** The number of packets in the sample */
  u_long packetsReceived;

  /** A guess as to the number of packets lost in the sample */
  u_long packetsLost;

  /** The average packet size for this sample */
  u_long packetSize;

  /** The smoothed round trip time for this sample, in microseconds. */
  u_long srtt;
  
private:

  /** Wire format of report. All multiple-byte fields are kept in network order */
  struct WireFormat{
    FlowIDStruct id;
    u_long mostSigNTP;
    u_long leastSigNTP;
    u_long sampLen;
    u_long sampDur;
    u_long packetsReceived;
    u_long packetsLost;
    u_long packetSize;
    u_long srtt;
  };
};

/** A Performance Request Object, usually used with a "spand.GetPerf"
    handler.
 * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.19 $.
 */
class PerfRequestPayload : public SendableSpandObject{
public:
  /** Default Constructor */
  PerfRequestPayload();

  /** Constructor that provides address to ask about */
  PerfRequestPayload(Addr addr0);

  /** Package this object into a sequence of bytes. The number of
      bytes written is returned. If wireData is NULL, then no bytes
      are actually written, but the size of the marshalled object is
      returned. */
  u_long Marshal(u_char *wireData);

  /** Unpackage this object from a sequence of bytes. The number of
      bytes read is returned. */
  u_long Unmarshal(u_char *wireData);

  /** Return a String representation of this object. */
  char * Print();
  
  /** The address of which to ask the performance */
  Addr addr;

private:
  /** Initializes fields */
  Init();

  /** Wire format of PerfRequestPayload message. All multi-byte fields
      are maintained in network order */
  struct WireFormat{
    struct in_addr addr;
  };

};

/** A Performance Response Message object, returned by a "spand.GetPerf" handler.
  * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.19 $.
 */
class PerfResponsePayload : public SendableSpandObject{
public:

  /** Default Constructor */
  PerfResponsePayload();

  /** Package this object into a sequence of bytes. The number of
      bytes written is returned. If wireData is NULL, then no bytes
      are actually written, but the size of the marshalled object is
      returned. */
  u_long Marshal(u_char *wireData);

  /** Unpackage this object from a sequence of bytes. The number of
      bytes read is returned. */
  u_long Unmarshal(u_char *wireData);

  /** Return a String representation of this object. */
  char * Print();
  
  /** The address that this response deals with */
  Addr addr;

  /** The average available bandwidth from this host, in bits/sec */
  u_long availBandwidth;

  /** Standard Deviation of available bandwidth from this host, in bits/sec */
  u_long stdDevAvailBandwidth;

  /** The probability that a packet will be lost along the path from this host, in 1/1000s */
  u_long packetLossProb;

  /** Standard Deviation of packet loss probability, in 1/1000s */
  u_long stdDevPacketLossProb;

  /** The number of Reports used to make this Response */
  u_long numReports;

  /** The total number of bytes seen from this host */
  u_long numBytes;

  /** The average Srtt from this host, in microseconds */
  u_long avgSrtt;

  /** Standard Deviation of srtt */
  u_long stdDevSrtt;
  
private:
  /** Initializes fields */
  Init();

  /** Wire format of PerfResponse message. All multi-byte fields are
      maintained in network order */
  struct WireFormat {
    struct in_addr addr;
    u_long availBandwidth;
    u_long stdDevAvailBandwidth;
    u_long packetLossProb;
    u_long stdDevPacketLossProb;
    u_long numReports;
    u_long numBytes;
    u_long avgSrtt;
    u_long stdDevSrtt;
  };
};

/** A Summary request asks for a summary of the number of reports
    received for a particular host. The object is exactly like a
    PerfRequest Object
 * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.19 $.
 */
class SummRequestPayload : public PerfRequestPayload {
};

/** A Summ. Response Message object contains the number of reports
    received for a particular host.
 * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.19 $.
 */
class SummResponsePayload : public SendableSpandObject{
public:

  /** Default Constructor */
  SummResponsePayload();

  /** Constructor from explicit fields */
  SummResponsePayload(Addr addr0, u_long numReports);

  /** Package this object into a sequence of bytes. The number of
      bytes written is returned. If wireData is NULL, then no bytes
      are actually written, but the size of the marshalled object is
      returned. */
  u_long Marshal(u_char *wireData);

  /** Unpackage this object from a sequence of bytes. The number of
      bytes read is returned. */
  u_long Unmarshal(u_char *wireData);

  /** Return a String representation of this object. */
  char * Print();
  
  /** The source addr that this message is about */
  Addr addr;

  /** The number of reports seen for this address */
  u_long numReports;
  
private:
  /** Initializes fields */
  Init();

  /** Wire format of PerfResponse message. All multi-byte fields are
      maintained in network order */
  struct WireFormat {
    struct in_addr addr;
    u_long numReports;
  };
};

/** A Raw Request message asks for all Reports for a given address. It
    is exactly the same as a PerfRequestPayload
 * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.19 $.
 */
class RawRequestPayload : public PerfRequestPayload {
};

/** A RawResponsePayload is exactly the same as a ReportPayload
  * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.19 $.
 */
class RawResponsePayload: public ReportPayload {
};

/** An ErrorPayload is used to return error messages. It just contains
  * an error string.
  * @version $Revision: 1.19 $.
  * @author SPAND Group, UC Berkeley.
  */
class ErrorPayload: public SendableSpandObject {
 public:

  /** The actual error message */
  char *errorMsg;

  /** Default constructor */
  ErrorPayload();

  /** Constructor from explicit message */
  ErrorPayload(char *errorMsg0);

  /** Descructor (frees errorMsg space) */
  ~ErrorPayload();

  /** Package this object into a sequence of bytes. The number of
      bytes written is returned. If wireData is NULL, then no bytes
      are actually written, but the size of the marshalled object is
      returned. */
  u_long Marshal(u_char *wireData);

  /** Unpackage this object from a sequence of bytes. The number of
      bytes read is returned. */
  u_long Unmarshal(u_char *wireData);

  /** Return a String representation of this object. */
  char * Print();

};

/** A Spand message consists of a marshalled data object and a
 * handler. The handler either indicates the type of the marshalled
 * object (if active=0) or the name of the handler of the marshalled
 * data (if active=1).
 * @version $Revision: 1.19 $
 * @author SPAND Group, UC Berkeley. 
 */
class Msg: public SendableSpandObject{
public:

  /** The name of the handler. Stored as a fixed string for simplicity */
   char handler[MAX_MSG_SIZE];

   /** Whether the handler string indicates the handler or indicates
      the type of the payload */
   int active;

   /** The length of the marshalled object */
  int dataLen;

  /** A pointed to the marshalled object data */
  unsigned char *data;

  /** Default Constuctor */
  Msg() {};

  /** Destructor. Frees data pointer */
  ~Msg();

  /** Constructor from socket. Reads data from socket and unmarshalls Msg object */
  Msg(Socket *readFrom);

  /** Constructor from Datagram Socket */
  Msg(UdpSocket *readFrom);

  /** Constructor from Stream Socket */
  Msg(TcpSocket *readFrom);

  /** Constructor from handler and payload */
  Msg(char *handler0, SendableSpandObject &o, int active0);

  /** Create the Message. Creates an object specified by the handler
      type and unmarshalls the data */
  SendableSpandObject *ToObject();

    /** Package this object into a sequence of bytes. The number of
      bytes written is returned. If wireData is NULL, then no bytes
      are actually written, but the size of the marshalled object is
      returned. */
  u_long Marshal(u_char *wireData);

  /** Unpackage this object from a sequence of bytes. The number of
      bytes read is returned. */
  u_long Unmarshal(u_char *wireData);

  /** Return a String representation of this object. */
  char * Print();
  
  /** An empty message (no contents). Useful in determining whether or
     not a socket read read anything */
  int empty;

 private:
  /** Helper function that reads from socket and checks for errors */
  void ReadSock(Socket *readFrom, char *buf, int size);
  
  /** Helper function that reads from socket and checks for errors */
  void ReadFullSock(Socket *readFrom, char *buf, int size);

};

/** A Flow contains a ReportPayload struct and has additional
    variables for start/end of flow, etc.
 * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.19 $.
 */
class Flow :  public ReportPayload {
public:
  /** Default Constructor */
  Flow();

  /** Constructor from Flow ID */
  Flow(FlowID id0):
    ReportPayload(id0)
    { startSeq=0; startTime.tv_sec=0; startTime.tv_usec=0;}

  /** A constructor from wire Data */
  Flow (u_char *wireData) { Unmarshal(wireData);};

  /** Package this object into a sequence of bytes. The number of
      bytes written is returned. If wireData is NULL, then no bytes
      are actually written, but the size of the marshalled object is
      returned. */
  u_long Marshal(u_char *wireData);

  /** Unpackage this object from a sequence of bytes. The number of
      bytes read is returned. */
  u_long Unmarshal(u_char *wireData);

  /** Starting TCP sequence number of the connection */
  u_long startSeq;

  /** Last TCP sequence number seen */
  u_long lastSeq;

  /** Maximum TCP sequence number seen so far */
  u_long maxSeq;

  /** The sequence number that marks the beginning of the current RTT measurement. */
  u_long rttSeq;

  /** Time at which flow was first seen */
  struct timeval startTime;

  /** Last time a packet was seen on this flow */
  struct timeval lastPktTime;

  /** The timestamp of the start of the current RTT measurement. */
  struct timeval rttTs;

private:

  /** Initialize Flow struct */
  void Init(FlowID id0);

  /** Wire format of report. All multiple-byte fields are kept in
      network order */
  struct WireFormat{
    u_long startSeq;
    u_long lastSeq;
    struct timeval startTime;
  };
};

/** A Requestor object sends a message to the perf server and waits
    for a reply. It tries to contact the Perf Server using UDP and
    falls back to TCP if the reply is too large to fit in a single
    datagram or if there is no reponse using TCP.
 * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.19 $.
 */
class Requestor: public Spand {
public:

  /** Default constructor. It must take the name of Perf Server as an
     argument */
  Requestor(char *servername0);

  /** Default destructor. It closes sockets used to communicate with
      the Perf. Server */
  ~Requestor();
  
  /** Make a request and receive a Message as a reply (if waitForReply
      is 1) */
  Msg *Request(Msg *m, int waitForReply);

  
private:

  /** Connect to the Perf Server using a Stream Socket */
  ConnectStreamSpand();

  /** Connect to the Perf Server using a Datagram Socket */
  ConnectDgramSpand();

  /** Send a Message to the Perf Server using a Stream Socket */
  Msg *RequestStream(Msg &m, int waitForReply);

  /** Send a Message to the Perf Server using a Datagram Socket */
  Msg *RequestDgram(Msg &m, int waitForReply);

  /** The name of the Perf Server */
  char servername[MAX_MSG_SIZE];

  /** The socket used to communicate with the Perf Server */
  Socket *requestSock;
};

/** A collection of SendableSpandObjects.
  * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.19 $.
 */
class SendableSpandObjectColl: public SendableSpandObject {

 public:
  
  /** Package this object into a sequence of bytes. The number of
      bytes written is returned. If wireData is NULL, then no bytes
      are actually written, but the size of the marshalled object is
      returned. */
  u_long Marshal(u_char *wireData);

  /** Unpackage this object from a sequence of bytes. The number of
      bytes read is returned. */
  u_long Unmarshal(u_char *wireData);

  /** Return a String representation of this object. */
  char * Print();
  
  /** The Collection of objects */
  Collection<SendableSpandObject *> objects;

  /** The type of the objects in the collection */
  char name[MAX_MSG_SIZE]; 

  /** Default constructor */
  SendableSpandObjectColl();

  /** Default destructor */
  /*  ~SendableSpandObjectColl();*/

 protected:

  /** Used for printing the collection of objects. It is not a fixed
      size so it can deal with very large collections of objects */
  char *printBuf;
};

/** This object plays a role similar to Java's
    Class.getByName(..).newInstance(). Given the name of an object, it
    creates an object of that type and returns it.
 * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.19 $.
 */
class SpandObjectFactory: public Spand {
 public:
  /** Create an object of type name and return it */
  SendableSpandObject *MakeObject(char *name);
};
















