/*
 * Copyright (c) 1996-1998 Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the SPAND Research
 *      Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/* $Header: /home/stemm/cvs/option/option.h,v 1.4 1998/06/10 23:23:11 stemm Exp $ */
/*
 * option.h --
 *	This defines the Option type and the interface to the
 *	Opt_Parse library call that parses command lines.
 */

#ifndef _OPTION
#define _OPTION

/** @name option library
    This library is used for parsing of command line arguments.
 * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.4 $.
 */

//@{

/** An array of option descriptions (type Option) is passed into the routine which interprets the command line. Each option description includes the key-string that indicates the option, a type for the option, the address of an associated variable, and a documentation message that is printed when the command is invoked with a single argument of '?'.  */

struct _Option {
  /** Indicates option type, see below. */
  int type;
  
  /** The key string that flags option. */
  char *key;

  /** Address of variable to modify. */
  char *address;
  
  /** Documentation message. */
  char *docMsg;
};

typedef struct _Option Option;

#define	OPT_BASE_VALUE		-1000

/** If the flag is present then set the associated (integer) variable to TRUE (1).  */
#define OPT_TRUE 1

/** If the flag is present then set the associated (integer) variable to val. Val must be a non-negative integer.*/
#define OPT_CONSTANT(val) ((int) val)

/** If the flag is present then set the associated (integer) variable to FALSE (0).  */
#define OPT_FALSE 0

/** If the flag is present then the next argument on the command line
    is interpreted as an integer and that value is assigned to the
    options associated variable.  */
#define OPT_INT			(OPT_BASE_VALUE -1)

/** If the flag is present then the next argument on the command line is copied into the string variable associated with the option.  */
#define OPT_STRING		(OPT_BASE_VALUE -2)

/** If the flag is present, inhibit processing of later options, so
    that they're all returned to the caller in argv.  In addition, set
    the associated variable to the index of the first of these
    arguments in the returned argv.  This permits a program to allow a
    flag to separate its own options from options it will pass to
    another program. */
#define OPT_REST		(OPT_BASE_VALUE -3)

/** If the flag is present then the next argument on the
   command line is interpreted as a "double" and that value is
   assigned to the option's associated variable.  */
#define OPT_FLOAT		(OPT_BASE_VALUE -4)

/** If the flag is present, pass the next argument to
"address" as a function. The function should be declared:
\begin{verbatim}
    int func(optString, arg)
	char 	*optString;
	char	*arg;
\end{verbatim}
   	Func should return non-zero if the argument was consumed or
   	zero if not.  "optString" is the option key string that caused
   	the function to be called and "arg" is the next argument (if
   	there is no next argument then "arg" will be NULL). */
#define OPT_FUNC		(OPT_BASE_VALUE -5)

/** If the flag is present, pass the remaining arguments and the
number of arguments to "address" as a function. The function should be
declared:
\begin{verbatim}
    int
    func(optString, argc, argv)
	char *optString;
	int argc;
	char **argv;
\end{verbatim}
and should return the new number of arguments left in argv.  argv
should have been shuffled to eliminate the arguments func consumed. */
#define OPT_GENFUNC		(OPT_BASE_VALUE -6)

/** A dummy entry. Exists mostly for its documentation string.  As an
 additional side effect, if its key string an argument, Opt_Parse will
 treat it like a question mark (i.e. print out the program's usage and
 exit).  */
#define OPT_DOC			(OPT_BASE_VALUE -7)


/** If the flag is present then the next argument on the command line
    is interpreted as a date and time.  The corresponding time value
    (number of seconds past the epoch) is assigned to the option's
    associated variable. */
#define OPT_TIME		(OPT_BASE_VALUE -8)

/** If the flag is present then the next argument on the command line
    is interpreted as a hexidecimal integer and that value is assigned
    to the options associated variable. (which is an int). */
#define OPT_HEX			(OPT_BASE_VALUE -9)

/** If the flag is present then the next argument on the command line
    is interpreted as a size (which understands shortcuts such as
    K=1000, etc.) and and that value is assigned to the options
    associated variable (which is an int).  */
#define OPT_SIZE			(OPT_BASE_VALUE -10)

/** If the flag is present then the next argument on the command line
    is interpreted as a size (which understands shortcuts such as
    K=1000, etc.) and and that value is assigned to the options
    associated variable (which is an double).  */
#define OPT_SIZE_DOUBLE			(OPT_BASE_VALUE -11)

/*
 * Flag values for Opt_Parse:
 */
 
/** Permit many flags to be clustered under a single "-".  In
    otherwords, treat "foo -abc" the same as "foo -a -b -c". */
#define OPT_ALLOW_CLUSTERING	1

/** Stop parsing if something other than an option (starting with a hyphen) is encountered.  */
#define OPT_OPTIONS_FIRST	2

/** Do not consider case when checking flags.  */
#define OPT_CASE_INSENSITIVE	0x10

/*
 * Exported procedures:
 */

#ifndef __cplusplus
/** Process a command line according to a template of accepted options.

 * {\bf Results:} The number of options that weren't processed by this procedure
 *	is returned, and argv points to an array of unprocessed
 *	options.  (This is all of the options that didn't start with
 *	"-", except for those used as arguments to the options
 *	processed here; it's also anything after an OPT_REST option.)
 *
 * {\bf Side effects:} The variables referenced from the option array get modified
 *	if their option was present on the command line.  Can clobber 
 *	the global buffer used by localtime(3).
 */
int 	Opt_Parse (int argc, char **argv, Option *optionArray, int numOptions, int flags);

/** Print out a usage message for a command.  This prints out the
 *	documentation strings associated with each option.
 *
 * {\bf Results:} none.
 *
 * {\bf Side effects:} Messages printed onto the console.
 */
void 	Opt_PrintUsage (char* commandName, Option optionArray[], int numOptions);
#else

extern "C" int	Opt_Parse (int argc, char** argv, Option optionArray[], int numOptions, int flags);
extern "C" void	Opt_PrintUsage (char* commandName, Option optionArray[], int numOptions);
#endif

/**
 * Macro to determine size of option array.  
 */

#define Opt_Number(optionArray)	(sizeof(optionArray)/sizeof((optionArray)[0]))

//@}
#endif /* _OPTION */
