/*
 * Copyright (c) 1996-1998 Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the SPAND Research
 *      Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
/* $Header: /home/stemm/cvs/SPAND/src/snooper/snooper.cc,v 1.12 1998/06/11 00:07:24 stemm Exp $ */
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/select.h>
#include <math.h>
#include <strings.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>

#include "porting.h"
#include "Socket.h"
extern "C" {
#include "dbgprintlib.h"
#include "option.h"
}
#include "handlers.h"
#include <Spand.h>

extern "C" {
#include <pcap.h>
}

/** @name snooper program
    snooper acts as a packet capture host, snooping on local area network traffic and generating performance reports on behalf of clients. The usage is {\tt snooper [-spandSvr <addr>] [-warmThresh <\# reports>] [-minLen <\# bytes>] [-test] [-r <filename>] [-i <iface>] [-help]}.

    {\tt -spandSvr <addr>}: Name/Address of Spand Perf Server. The default value is 127.0.0.1.
    
    {\tt -warmThresh <\# reports>}: When used with {\tt -test}, specifies the number of reports that must be generated for a host before querying about the performance for that host. The default is 0. 

    {\tt -minLen <\# bytes>}: Minimum transfer length to report. The default is 1000. 

    {\tt -test}: Test the performance server by generating a Performance Request for a host immediately before generating a performance Report for that host. This can be useful in testing the accuracy of Performance Responses. The default is false (do not test). 

    {\tt -r <filename>}: Read packets from file {\tt <filename>}. The file is in libpcap format. 
    
    {\tt -i <iface>}: Interface to snoop on. 
    
    {\tt -help}: Print a summary of the command line arguments.
     * @author SPAND Group, UC Berkeley.
 * @version $Revision: 1.12 $.
*/

char *spandSvr="127.0.0.1";
int warmThresh=0;
int minLen=1000;
int test_aggregator=0;
char *readFile=NULL;
char *device=NULL;

Option optionArray[] = {
  {OPT_STRING, "spandSvr", (char *) &spandSvr, "Name/Address of Spand Perf Server"},
  {OPT_INT, "warmThresh", (char *) &warmThresh, "Number of reports to wait for warm system before querying"},
  {OPT_INT, "minLen", (char *) &minLen, "Minimum transfer length to report"},
  {OPT_TRUE, "test", (char *) &test_aggregator, "Test aggregator by making queries"},
  {OPT_STRING, "r", (char *) &readFile, "Input file to read"},
  {OPT_STRING, "i", (char *) &device, "Interface to snoop on"}
};

char *program_name="snooper";

/* Args: <snooper> client*/
void error(const char *fmt, ...)
{
	va_list ap;

	(void)fprintf(stderr, "%s: ", program_name);
	va_start(ap, fmt);
	(void)vfprintf(stderr, fmt, ap);
	va_end(ap);
	if (*fmt) {
		fmt += strlen(fmt);
		if (fmt[-1] != '\n')
			(void)fputc('\n', stderr);
	}
	exit(1);
	/* NOTREACHED */
}
struct handler {
	pcap_handler f;
	int type;
};

#if defined(AIX4)
#include <net/if_types.h>
static struct handler handlers[] = {
        { ether_if_handler,	IFT_ETHER},
	{ token_if_handler,	IFT_ISO88025},
	{ NULL,			0 },
};
#else
static struct handler handlers[] = {
	{ ether_if_handler,	DLT_EN10MB },
	{ NULL,			0 },
};
#endif	

static pcap_handler
lookup_ll_handler(int type)
{
	struct handler *p;

	for (p = handlers; p->f; ++p)
		if (type == p->type)
			return p->f;

	error("unknown data link type 0x%x", type);
	/* NOTREACHED */

	return NULL;
}

/*
 * Copy arg vector into a new buffer, concatenating arguments with spaces.
 */
char *
copy_argv(register char **argv)
{
	register char **p;
	register u_int len = 0;
	char *buf;
	char *src, *dst;

	p = argv;
	if (*p == 0)
		return 0;

	while (*p)
		len += strlen(*p++) + 1;

	buf = (char *)malloc(len);
	if (buf == NULL)
		error("copy_argv: malloc");

	p = argv;
	dst = buf;
	while ((src = *p++) != NULL) {
		while ((*dst++ = *src++) != '\0')
			;
		dst[-1] = ' ';
	}
	dst[-1] = '\0';

	return buf;
}

void SnoopReports(int argc, char *argv[])
{
  u_char *pcap_userdata;
  char ebuf[PCAP_ERRBUF_SIZE];
  int snaplen=500;
  static pcap_t *pd;
  bpf_u_int32 localnet, netmask;
  int i;
  struct bpf_program fcode;
  char *cmdbuf;
  pcap_handler handler;
  int pcap_program_index = 1;
  
  cmdbuf = copy_argv(&argv[pcap_program_index]);

  if (device==NULL){
    device = pcap_lookupdev(ebuf);
    if (device==NULL){
      error("%s", ebuf);
    }
  }

  if (readFile != NULL) {
    setuid(getuid());
    pd =pcap_open_offline(readFile, ebuf);
    if (pd == NULL){
      error("%s", ebuf);
    }
    localnet = 0;
    netmask = 0;
  } else {
	pd = pcap_open_live(device, snaplen, 1, 1000, ebuf);
	if (pd == NULL){
	  error(ebuf);
	}
	i = pcap_snapshot(pd);
	if (snaplen < i) {
	  fprintf(stderr, "snaplen raised from %d to %d", snaplen, i);
	  snaplen = i;
	}
  
	if (pcap_lookupnet(device, &localnet, &netmask, ebuf) < 0)
	  error(ebuf);

	setuid(getuid());
  }
  
  if (pcap_compile(pd, &fcode, cmdbuf, 1, netmask)<0)
    error(pcap_geterr(pd));

  if (pcap_setfilter(pd, &fcode) < 0)
    error(pcap_geterr(pd));

  handler = lookup_ll_handler(pcap_datalink(pd));
  pcap_userdata = 0;
  
  if (pcap_loop(pd, -1, handler, pcap_userdata) < 0) {
    (void)fprintf(stderr, "%s: pcap_loop: %s\n",
		  program_name, pcap_geterr(pd));
    exit(1);
  }
  pcap_close(pd);
  exit(0);

}

main(int argc, char *argv[])
{
  fd_set fds;

  argc=Opt_Parse(argc, argv, optionArray, Opt_Number(optionArray), 0);

  /* Snoop on local traffic and generate reports */
  SnoopReports(argc, argv);

}



  
